package cache_test

import (
	"crypto/rand"
	"testing"

	"github.com/pkt-cash/pktd/btcutil"
	"github.com/pkt-cash/pktd/btcutil/gcs"
	"github.com/pkt-cash/pktd/chaincfg/chainhash"
	"github.com/pkt-cash/pktd/neutrino/cache"
	"github.com/pkt-cash/pktd/neutrino/cache/lru"
	"github.com/pkt-cash/pktd/neutrino/filterdb"
	"github.com/pkt-cash/pktd/wire"
)

// TestBlockFilterCaches tests that we can put and retrieve elements from all
// implementations of the filter and block caches.
func TestBlockFilterCaches(t *testing.T) {
	t.Parallel()

	const filterType = filterdb.RegularFilter

	// Create a cache large enough to not evict any item. We do this so we
	// don't have to worry about the eviction strategy of the tested
	// caches.
	const numElements = 10
	const cacheSize = 100000

	// Initialize all types of caches we want to test, for both filters and
	// blocks. Currently the LRU cache is the only implementation.
	filterCaches := []cache.Cache{lru.NewCache(cacheSize)}
	blockCaches := []cache.Cache{lru.NewCache(cacheSize)}

	// Generate a list of hashes, filters and blocks that we will use as
	// cache keys an values.
	var (
		blockHashes []chainhash.Hash
		filters     []*gcs.Filter
		blocks      []*btcutil.Block
	)
	for i := 0; i < numElements; i++ {
		var blockHash chainhash.Hash
		if _, err := rand.Read(blockHash[:]); err != nil {
			t.Fatalf("unable to read rand: %v", err)
		}

		blockHashes = append(blockHashes, blockHash)

		filter, err := gcs.FromBytes(
			uint32(i), uint8(i), uint64(i), []byte{byte(i)},
		)
		if err != nil {
			t.Fatalf("unable to create filter: %v", err)
		}
		filters = append(filters, filter)

		// Put the generated filter in the filter caches.
		cacheKey := cache.FilterCacheKey{blockHash, filterType}
		for _, c := range filterCaches {
			c.Put(cacheKey, &cache.CacheableFilter{Filter: filter})
		}

		msgBlock := &wire.MsgBlock{}
		block := btcutil.NewBlock(msgBlock)
		blocks = append(blocks, block)

		// Add the block to the block caches, using the block INV
		// vector as key.
		blockKey := wire.NewInvVect(
			wire.InvTypeWitnessBlock, &blockHash,
		)
		for _, c := range blockCaches {
			c.Put(*blockKey, &cache.CacheableBlock{block})
		}
	}

	// Now go through the list of block hashes, and make sure we can
	// retrieve all elements from the caches.
	for i, blockHash := range blockHashes {
		// Check filter caches.
		cacheKey := cache.FilterCacheKey{blockHash, filterType}
		for _, c := range filterCaches {
			e, err := c.Get(cacheKey)
			if err != nil {
				t.Fatalf("Unable to get filter: %v", err)
			}

			// Ensure we got the correct filter.
			filter := e.(*cache.CacheableFilter).Filter
			if filter != filters[i] {
				t.Fatalf("Filters not equal: %v vs %v ",
					filter, filters[i])
			}
		}

		// Check block caches.
		blockKey := wire.NewInvVect(
			wire.InvTypeWitnessBlock, &blockHash,
		)
		for _, c := range blockCaches {
			b, err := c.Get(*blockKey)
			if err != nil {
				t.Fatalf("Unable to get block: %v", err)
			}

			// Ensure it is the same block.
			block := b.(*cache.CacheableBlock).Block
			if block != blocks[i] {
				t.Fatalf("Not equal: %v vs %v ",
					block, blocks[i])
			}
		}
	}
}
